import unittest

from factorio_circuit_ast.DoubleNetwork import *
from factorio_circuit_ast.Signal import Signals


class TestDoubleNetwork(unittest.TestCase):
    def test_constructor(self):
        d_net = DoubleNetwork()
        self.assertIsNone(d_net.red)
        self.assertIsNone(d_net.green)
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        d_net = DoubleNetwork(r_net, g_net)
        self.assertIs(d_net.red, r_net)
        self.assertIs(d_net.green, g_net)
        with self.assertRaises(TypeError):
            DoubleNetwork(red=g_net)
        with self.assertRaises(TypeError):
            DoubleNetwork(green=r_net)

    def test_set_red(self):
        d_net = DoubleNetwork()
        with self.assertRaises(TypeError):
            d_net.red = Network(network=NetworkType.GREEN)
        r_net = Network(network=NetworkType.RED)
        d_net.red = r_net
        self.assertIs(d_net.red, r_net)
        d_net.red = None
        self.assertIsNone(d_net.red)

    def test_set_green(self):
        d_net = DoubleNetwork()
        with self.assertRaises(TypeError):
            d_net.green = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        d_net.green = g_net
        self.assertIs(d_net.green, g_net)
        d_net.green = None
        self.assertIsNone(d_net.green)

    def test_get_signal(self):
        d_net = DoubleNetwork()
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        with self.assertRaises(AssertionError):
            d_net.get_signal(Signal.SIGNAL_EACH)
        self.assertEqual(d_net.get_signal(Signal.SIGNAL_A), 0)
        d_net.red = r_net
        d_net.green = g_net
        self.assertEqual(d_net.get_signal(Signal.SIGNAL_A), 0)
        r_net.update_signal(Signal.SIGNAL_A, np.int32(10))
        g_net.update_signal(Signal.SIGNAL_A, np.int32(20))
        g_net.update_signal(Signal.SIGNAL_B, np.int32(100))
        r_net.tick()
        g_net.tick()
        self.assertEqual(d_net.get_signal(Signal.SIGNAL_A), 30)
        self.assertIs(type(d_net.get_signal(Signal.SIGNAL_A)), np.int32)
        self.assertEqual(d_net.get_signal(Signal.SIGNAL_B), 100)
        self.assertIs(type(d_net.get_signal(Signal.SIGNAL_B)), np.int32)

    def test_get_signals(self):
        d_net = DoubleNetwork()
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        self.assertEqual(d_net.get_signals(), {})
        d_net.red = r_net
        d_net.green = g_net
        self.assertEqual(d_net.get_signals(), {})
        r_net.update_signal(Signal.SIGNAL_A, np.int32(10))
        g_net.update_signal(Signal.SIGNAL_A, np.int32(20))
        g_net.update_signal(Signal.SIGNAL_B, np.int32(100))
        r_net.tick()
        g_net.tick()
        self.assertEqual(
            d_net.get_signals(), {Signal.SIGNAL_A: 30, Signal.SIGNAL_B: 100}
        )
        self.assertTrue(
            all(isinstance(i, np.int32) for i in d_net.get_signals().values())
        )

    def test_update_signal(self):
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        d_net = DoubleNetwork(r_net, g_net)
        with self.assertRaises(AssertionError):
            d_net.update_signal(Signal.SIGNAL_EACH, np.int32(10))
        d_net.update_signal(Signal.ACCUMULATOR, np.int32(10))
        r_net.tick()
        g_net.tick()
        self.assertEqual(r_net.get_signal(Signal.ACCUMULATOR), 10)
        self.assertEqual(g_net.get_signal(Signal.ACCUMULATOR), 10)
        d_net.red = None
        d_net.update_signal(Signal.SIGNAL_A, np.int32(30))
        g_net.tick()
        self.assertEqual(g_net.get_signal(Signal.SIGNAL_A), 30)
        d_net.red = r_net
        d_net.green = None
        d_net.update_signal(Signal.SIGNAL_A, np.int32(50))
        r_net.tick()
        self.assertEqual(r_net.get_signal(Signal.SIGNAL_A), 50)

    def test_update_signals(self):
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        d_net = DoubleNetwork(r_net, g_net)
        with self.assertRaises(AssertionError):
            d_net.update_signals({Signal.SIGNAL_EACH: np.int32(10)})
        signals: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_B: 30,
        }
        d_net.update_signals(signals)
        r_net.tick()
        g_net.tick()
        self.assertEqual(r_net.get_signals(), signals)
        self.assertEqual(g_net.get_signals(), signals)
        d_net.red = None
        d_net.update_signals(signals)
        g_net.tick()
        self.assertEqual(g_net.get_signals(), signals)
        d_net.red = r_net
        d_net.green = None
        d_net.update_signals(signals)
        r_net.tick()
        self.assertEqual(r_net.get_signals(), signals)

    def test_eq(self):
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        net1 = DoubleNetwork()
        net2 = DoubleNetwork()
        self.assertNotEqual(net1, r_net)
        self.assertEqual(net1, net2)
        net1.red = r_net
        self.assertNotEqual(net1, net2)
        net2.red = r_net
        self.assertEqual(net1, net2)
        net2.green = g_net
        self.assertNotEqual(net1, net2)
        net1.green = g_net
        self.assertEqual(net1, net2)

    def test_cli_repr(self):
        r_net = Network(network=NetworkType.RED)
        g_net = Network(network=NetworkType.GREEN)
        net = DoubleNetwork()
        self.assertEqual(net.cli_repr(), "[r=-,g=-]")
        net.red = r_net
        self.assertEqual(net.cli_repr(), f"[r=n{r_net.id},g=-]")
        net.green = g_net
        self.assertEqual(net.cli_repr(), f"[r=n{r_net.id},g=n{g_net.id}]")



import unittest

import numpy as np

from factorio_circuit_ast.Constant import *
from factorio_circuit_ast.Network import Network, NetworkType
from factorio_circuit_ast.Signal import Signal


def get_nets() -> DoubleNetwork:
    r_net = Network(network=NetworkType.RED)
    g_net = Network(network=NetworkType.GREEN)
    d_net = DoubleNetwork(r_net, g_net)
    return d_net


class TestConstant(unittest.TestCase):
    def test_constructor(self):
        net = get_nets()
        const = Constant(net, {Signal.SIGNAL_A: np.int32(10)}, False)
        self.assertEqual(const.is_on, False)
        self.assertEqual(const.output_network, net)
        self.assertEqual(const.signals, {Signal.SIGNAL_A: 10})
        const = Constant()
        self.assertEqual(const.is_on, True)
        self.assertEqual(const.output_network, None)
        self.assertEqual(const.signals, {})

    def test_signals_setter(self):
        with self.assertRaises(AssertionError):
            Constant(signals={Signal.SIGNAL_EVERYTHING: np.int32(10)})

    def test_tick_on(self):
        net = get_nets()
        sig: Signals = {
            Signal.SIGNAL_A: np.int32(10),
            Signal.RAIL_SIGNAL: np.int32(20),
        }
        const = Constant(net, sig)
        const.tick()
        self.assertEqual(const.previous_result, sig)
        net.red.tick()
        self.assertEqual(net.get_signals(), sig)
        const.output_network = None
        const.tick()
        self.assertEqual(const.previous_result, sig)
        const.is_on = False
        const.tick()
        self.assertEqual(const.previous_result, {})

    def test_tick_off(self):
        net = get_nets()
        sig: Signals = {
            Signal.SIGNAL_A: np.int32(10),
            Signal.RAIL_SIGNAL: np.int32(20),
        }
        const = Constant(net, sig, False)
        const.tick()
        self.assertEqual(const.previous_result, {})
        net.red.tick()
        self.assertEqual(net.get_signals(), {})

    def test_cli_repr(self):
        net = get_nets()
        sig: Signals = {
            Signal.SIGNAL_A: np.int32(10),
            Signal.RAIL_SIGNAL: np.int32(20),
        }
        const = Constant(net, sig, False)
        self.assertEqual(
            const.cli_repr(),
            f"Constant output={net.cli_repr()} [signal-A: 10, rail-signal: 20] (off)",
        )
        const = Constant()
        self.assertEqual(const.cli_repr(), f"Constant output=- [] (on)")



import unittest

from factorio_circuit_ast.Network import *
from factorio_circuit_ast.Signal import Signals


class TestNetwork(unittest.TestCase):
    def test_constructor(self):
        net = Network()
        self.assertEqual(net._depends, [])
        self.assertEqual(net._dependants, [])
        self.assertEqual(net._previous_state, {})
        self.assertEqual(net._current_state, {})
        self.assertIsInstance(net.id, int)
        self.assertIs(networks[net.id], net)
        self.assertEqual(net.type, NetworkType.RED)
        net = Network(nid=1024, network=NetworkType.GREEN)
        self.assertEqual(net.id, 1024)
        self.assertIs(networks[1024], net)
        self.assertEqual(net.type, NetworkType.GREEN)

    def test_get_signal(self):
        net = Network()
        self.assertEqual(net.get_signal(Signal.ADVANCED_CIRCUIT), 0)
        with self.assertRaises(AssertionError):
            net.get_signal(Signal.SIGNAL_ANYTHING)

    def test_tick(self):
        net = Network()
        net._current_state[Signal.ADVANCED_CIRCUIT] = np.int32(10)
        net.tick()
        self.assertEqual(net.get_signal(Signal.ADVANCED_CIRCUIT), 10)

    def test_update_signal(self):
        net = Network()
        net.update_signal(Signal.SIGNAL_A, np.int32(10))
        with self.assertRaises(AssertionError):
            net.update_signal(Signal.SIGNAL_EVERYTHING, np.int32(0))
        self.assertEqual(net.get_signal(Signal.SIGNAL_A), 0)
        self.assertEqual(net._current_state[Signal.SIGNAL_A], 10)
        net.tick()
        self.assertEqual(net.get_signal(Signal.SIGNAL_A), 10)
        self.assertNotIn(Signal.SIGNAL_A, net._current_state)
        net.update_signal(Signal.SIGNAL_A, np.int32(10))
        self.assertEqual(net._current_state[Signal.SIGNAL_A], 10)
        net.update_signal(Signal.SIGNAL_A, np.int32(20))
        self.assertEqual(net._current_state[Signal.SIGNAL_A], 30)
        net.update_signal(Signal.ADVANCED_CIRCUIT, np.int32(500))
        self.assertEqual(net._current_state[Signal.SIGNAL_A], 30)
        self.assertEqual(net._current_state[Signal.ADVANCED_CIRCUIT], 500)
        net.tick()
        self.assertEqual(net.get_signal(Signal.SIGNAL_A), 30)
        self.assertEqual(net.get_signal(Signal.ADVANCED_CIRCUIT), 500)
        net.update_signal(Signal.SIGNAL_A, np.int32(2**31 - 1))
        net.update_signal(Signal.SIGNAL_A, np.int32(1))
        net.tick()
        self.assertEqual(net.get_signal(Signal.SIGNAL_A), -(2**31))
        self.assertIsInstance(net.get_signal(Signal.SIGNAL_A), np.int32)

    def test_get_signals(self):
        net = Network()
        self.assertEqual(net.get_signals(), {})
        net.update_signal(Signal.SIGNAL_A, np.int32(10))
        net.update_signal(Signal.SIGNAL_B, np.int32(30))
        self.assertEqual(net.get_signals(), {})
        net.tick()
        self.assertEqual(net.get_signals(), {Signal.SIGNAL_A: 10, Signal.SIGNAL_B: 30})
        self.assertTrue(
            all(isinstance(i, np.int32) for i in net.get_signals().values())
        )
        net.tick()
        self.assertEqual(net.get_signals(), {})

    def test_update_signals(self):
        net = Network()
        with self.assertRaises(AssertionError):
            net.update_signals({Signal.SIGNAL_EACH: 10})
        with self.assertRaises(AssertionError):
            net.update_signals({Signal.SIGNAL_EVERYTHING: 10})
        with self.assertRaises(AssertionError):
            net.update_signals({Signal.SIGNAL_ANYTHING: 10})
        values: Signals = {
            Signal.SIGNAL_A: 20,
            Signal.SIGNAL_B: 50,
        }
        net.update_signals(values)
        self.assertEqual(net.get_signals(), {})
        net.tick()
        self.assertEqual(net.get_signals(), values)

    def test_cli_repr(self):
        net = Network()
        self.assertEqual(net.cli_repr(), f"n{net.id}")


class TestGetOrCreateNetwork(unittest.TestCase):
    def test(self):
        net = Network(2048, NetworkType.RED)
        self.assertIs(get_or_create_network(2048, NetworkType.RED), net)
        with self.assertRaises(AssertionError):
            get_or_create_network(2048, NetworkType.GREEN)
        self.assertIsInstance(get_or_create_network(2049, NetworkType.GREEN), Network)



import unittest

from factorio_circuit_ast.Network import Network, NetworkType
from factorio_circuit_ast.Operation import *


class TestNumericOperator(unittest.TestCase):
    def test_add(self):
        self.assertEqual(NumericOperator.ADD.calculate(np.int32(1), np.int32(2)), 3)
        self.assertEqual(
            NumericOperator.ADD.calculate(np.int32(2**31 - 1), np.int32(1)),
            -(2**31),
        )

    def test_subtract(self):
        self.assertEqual(
            NumericOperator.SUBTRACT.calculate(np.int32(2), np.int32(1)), 1
        )
        self.assertEqual(
            NumericOperator.SUBTRACT.calculate(np.int32(1), np.int32(2)), -1
        )

    def test_multiply(self):
        self.assertEqual(
            NumericOperator.MULTIPLY.calculate(np.int32(9), np.int32(9)), 81
        )

    def test_divide(self):
        self.assertEqual(NumericOperator.DIVIDE.calculate(np.int32(10), np.int32(2)), 5)
        self.assertEqual(
            NumericOperator.DIVIDE.calculate(np.int32(10), np.int32(11)), 0
        )

    def test_modulo(self):
        self.assertEqual(
            NumericOperator.MODULO.calculate(np.int32(10), np.int32(11)), 10
        )
        self.assertEqual(NumericOperator.MODULO.calculate(np.int32(10), np.int32(5)), 0)

    def test_exponent(self):
        self.assertEqual(
            NumericOperator.EXPONENT.calculate(np.int32(2), np.int32(31)), -(2**31)
        )

    def test_left_bit_shift(self):
        self.assertEqual(
            NumericOperator.LEFT_BIT_SHIFT.calculate(np.int32(8), np.int32(3)), 64
        )

    def test_right_bit_shift(self):
        self.assertEqual(
            NumericOperator.RIGHT_BIT_SHIFT.calculate(np.int32(8), np.int32(3)), 1
        )

    def test_and(self):
        self.assertEqual(NumericOperator.AND.calculate(np.int32(15), np.int32(22)), 6)

    def test_or(self):
        self.assertEqual(NumericOperator.OR.calculate(np.int32(15), np.int32(22)), 31)

    def test_xor(self):
        self.assertEqual(NumericOperator.XOR.calculate(np.int32(15), np.int32(22)), 25)


def get_nets() -> DoubleNetwork:
    r_net = Network(network=NetworkType.RED)
    g_net = Network(network=NetworkType.GREEN)
    d_net = DoubleNetwork(r_net, g_net)
    return d_net


class TestOperation(unittest.TestCase):
    def test_constructor(self):
        i_net = get_nets()
        o_net = get_nets()
        op = Operation(NumericOperator.OR, SignalOperand(Signal.SIGNAL_A))
        self.assertEqual(op.operation, NumericOperator.OR)
        self.assertEqual(op.left, ConstantOperand(np.int32(0)))
        self.assertEqual(op.right, ConstantOperand(np.int32(0)))
        self.assertEqual(op.result, SignalOperand(Signal.SIGNAL_A))
        self.assertIsNone(op.previous_result)
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_RED),
            o_net,
            i_net,
            SignalOperand(Signal.SIGNAL_EACH),
            SignalOperand(Signal.SIGNAL_CYAN),
        )
        self.assertEqual(op.left, SignalOperand(Signal.SIGNAL_EACH))
        self.assertEqual(op.right, SignalOperand(Signal.SIGNAL_CYAN))
        self.assertEqual(op.input_network, i_net)
        self.assertEqual(op.output_network, o_net)
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_RED),
            i_net,
            o_net,
            ConstantOperand(np.int32(10)),
            ConstantOperand(np.int32(20)),
        )
        self.assertEqual(op.left, ConstantOperand(np.int32(10)))
        self.assertEqual(op.right, ConstantOperand(np.int32(20)))

    def test_eq(self):
        i_net = get_nets()
        o_net = get_nets()
        op1 = Operation(NumericOperator.OR, SignalOperand(Signal.SIGNAL_A))
        op2 = Operation(NumericOperator.ADD, SignalOperand(Signal.SIGNAL_A))
        self.assertNotEqual(op1, ConstantOperand(np.int32(2)))
        self.assertNotEqual(op1, op2)
        op2.operation = NumericOperator.OR
        op1.left = ConstantOperand(np.int32(10))
        self.assertNotEqual(op1, op2)
        op1.left = ConstantOperand(np.int32(0))
        op2.right = ConstantOperand(np.int32(50))
        self.assertNotEqual(op1, op2)
        op2.right = ConstantOperand(np.int32(0))
        op1.result = SignalOperand(Signal.SIGNAL_B)
        self.assertNotEqual(op1, op2)
        op1.result = SignalOperand(Signal.SIGNAL_A)
        op2.output_network = o_net
        self.assertNotEqual(op1, op2)
        op1.output_network = o_net
        op1.input_network = i_net
        self.assertNotEqual(op1, op2)
        op2.input_network = i_net
        self.assertEqual(op1, op2)

    def test_set_left(self):
        op = Operation(NumericOperator.ADD, SignalOperand(Signal.SIGNAL_A))
        with self.assertRaises(AssertionError):
            op.left = SignalOperand(Signal.SIGNAL_A)
        i_net = get_nets()
        op.input_network = i_net
        op.right = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.left = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.left = SignalOperand(Signal.SIGNAL_ANYTHING)
        with self.assertRaises(AssertionError):
            op.left = SignalOperand(Signal.SIGNAL_EVERYTHING)
        op.left = SignalOperand(Signal.SIGNAL_A)
        op.right = SignalOperand(Signal.SIGNAL_A)
        op.left = SignalOperand(Signal.SIGNAL_EACH)
        op.right = ConstantOperand(np.int32(10))
        op.left = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.left = SignalOperand(Signal.SIGNAL_A)

    def test_right(self):
        op = Operation(NumericOperator.ADD, SignalOperand(Signal.SIGNAL_A))
        with self.assertRaises(AssertionError):
            op.right = SignalOperand(Signal.SIGNAL_A)
        i_net = get_nets()
        op.input_network = i_net
        op.left = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.right = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.right = SignalOperand(Signal.SIGNAL_ANYTHING)
        with self.assertRaises(AssertionError):
            op.right = SignalOperand(Signal.SIGNAL_EVERYTHING)
        op.right = SignalOperand(Signal.SIGNAL_A)
        op.left = SignalOperand(Signal.SIGNAL_A)
        op.right = SignalOperand(Signal.SIGNAL_EACH)
        op.left = ConstantOperand(np.int32(10))
        op.right = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_EACH)
        with self.assertRaises(AssertionError):
            op.right = SignalOperand(Signal.SIGNAL_A)

    def test_set_result(self):
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD, SignalOperand(Signal.SIGNAL_A), input_network=i_net
        )
        with self.assertRaises(AssertionError):
            op.result = SignalOperand(Signal.SIGNAL_EACH)
        op.right = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_A)
        op.right = SignalOperand(Signal.SIGNAL_A)
        op.left = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_EACH)
        op.result = SignalOperand(Signal.SIGNAL_A)
        op.left = SignalOperand(Signal.SIGNAL_A)
        with self.assertRaises(AssertionError):
            op.result = SignalOperand(Signal.SIGNAL_EVERYTHING)
        with self.assertRaises(AssertionError):
            op.result = SignalOperand(Signal.SIGNAL_ANYTHING)

    def test_tick_constant(self):
        o_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            left=ConstantOperand(np.int32(10)),
            right=ConstantOperand(np.int32(10)),
        )
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signal(Signal.SIGNAL_A), 20)
        self.assertEqual(op.previous_result, {Signal.SIGNAL_A: 20})

    def test_tick_left_each_and_output_each(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_EACH),
            o_net,
            i_net,
            SignalOperand(Signal.SIGNAL_EACH),
            ConstantOperand(np.int32(10)),
        )
        i_net.update_signal(Signal.WATER, np.int32(10))
        i_net.update_signal(Signal.SIGNAL_RED, np.int32(20))
        i_net.red.tick()
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signals(), {Signal.WATER: 20, Signal.SIGNAL_RED: 30})

    def test_tick_right_each_and_output_each(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_EACH),
            o_net,
            i_net,
            ConstantOperand(np.int32(10)),
            SignalOperand(Signal.SIGNAL_EACH),
        )
        i_net.update_signal(Signal.WATER, np.int32(10))
        i_net.update_signal(Signal.SIGNAL_RED, np.int32(20))
        i_net.red.tick()
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signals(), {Signal.WATER: 20, Signal.SIGNAL_RED: 30})

    def test_tick_no_each_and_output_each(self):
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_CYAN),
            left=ConstantOperand(np.int32(10)),
            right=ConstantOperand(np.int32(10)),
        )
        op.result.signal = Signal.SIGNAL_EACH
        with self.assertRaises(AssertionError):
            op.tick()

    def test_tick_left_each_and_output_no_each(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_PINK),
            o_net,
            i_net,
            SignalOperand(Signal.SIGNAL_EACH),
            ConstantOperand(np.int32(10)),
        )
        i_net.update_signal(Signal.WATER, np.int32(10))
        i_net.update_signal(Signal.SIGNAL_RED, np.int32(20))
        i_net.red.tick()
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signal(Signal.SIGNAL_PINK), 50)

    def test_tick_right_each_and_output_no_each(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_WHITE),
            o_net,
            i_net,
            ConstantOperand(np.int32(10)),
            SignalOperand(Signal.SIGNAL_EACH),
        )
        i_net.update_signal(Signal.WATER, np.int32(10))
        i_net.update_signal(Signal.SIGNAL_RED, np.int32(20))
        i_net.red.tick()
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signal(Signal.SIGNAL_WHITE), 50)

    def test_calculate_signals(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_WHITE),
            o_net,
            i_net,
            ConstantOperand(np.int32(10)),
            SignalOperand(Signal.WATER),
        )
        i_net.update_signal(Signal.WATER, np.int32(10))
        i_net.update_signal(Signal.SIGNAL_RED, np.int32(20))
        i_net.red.tick()
        op.tick()
        o_net.red.tick()
        self.assertEqual(o_net.get_signal(Signal.SIGNAL_WHITE), 20)

    def test_no_output_net(self):
        op = Operation(
            NumericOperator.ADD,
            SignalOperand(Signal.SIGNAL_A),
            left=ConstantOperand(np.int32(10)),
        )
        op.tick()
        self.assertEqual(op.previous_result, {Signal.SIGNAL_A: 10})

    def test_cli_repr(self):
        o_net = get_nets()
        i_net = get_nets()
        op = Operation(NumericOperator.ADD, SignalOperand(Signal.SIGNAL_A))
        self.assertEqual(
            op.cli_repr(), "Operation input=- output=- (0 + 0) -> signal-A"
        )
        op = Operation(
            NumericOperator.DIVIDE,
            SignalOperand(Signal.SIGNAL_WHITE),
            o_net,
            i_net,
            ConstantOperand(np.int32(10)),
            SignalOperand(Signal.WATER),
        )
        self.assertEqual(
            op.cli_repr(),
            f"Operation input={i_net.cli_repr()} output={o_net.cli_repr()} (10 / water) -> signal-white",
        )



from abc import ABC, abstractmethod
from typing import Any

import numpy as np

from .Signal import Signal, WildcardSignal


class Operand(ABC):
    @abstractmethod
    def wildcard(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_each(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_anything(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_everything(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def cli_repr(self) -> str:
        raise NotImplementedError


class SignalOperand(Operand):
    def __init__(self, signal: Signal):
        self.signal: Signal = signal

    def wildcard(self) -> bool:
        return self.signal in WildcardSignal

    def is_each(self) -> bool:
        return self.signal == Signal.SIGNAL_EACH

    def is_anything(self) -> bool:
        return self.signal == Signal.SIGNAL_ANYTHING

    def is_everything(self) -> bool:
        return self.signal == Signal.SIGNAL_EVERYTHING

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SignalOperand):
            return self.signal == other.signal
        return False

    def __repr__(self) -> str:
        return f"SignalOperand(signal={self.signal!r})"

    def cli_repr(self) -> str:
        return f"{self.signal.value}"


class ConstantOperand(Operand):
    def __init__(self, constant: np.int32):
        self.constant: np.int32 = constant

    def wildcard(self) -> bool:
        return False

    def is_each(self) -> bool:
        return False

    def is_anything(self) -> bool:
        return False

    def is_everything(self) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ConstantOperand):
            return self.constant == other.constant
        return False

    def __repr__(self) -> str:
        return f"ConstantOperand(constant={self.constant})"

    def cli_repr(self) -> str:
        return f"{self.constant}"



from __future__ import annotations

from enum import Enum
from typing import List, Dict

import numpy as np


class Signal(Enum):
    ROCKET_FUEL: str = "rocket-fuel"
    ADVANCED_CIRCUIT: str = "advanced-circuit"
    SULFURIC_ACID_BARREL: str = "sulfuric-acid-barrel"
    EXPLOSIVES: str = "explosives"
    IRON_ORE: str = "iron-ore"
    URANIUM_FUEL_CELL: str = "uranium-fuel-cell"
    WOODEN_CHEST: str = "wooden-chest"
    CONSTANT_COMBINATOR: str = "constant-combinator"
    LIGHT_OIL_BARREL: str = "light-oil-barrel"
    RED_WIRE: str = "red-wire"
    CONSTRUCTION_ROBOT: str = "construction-robot"
    STEAM_TURBINE: str = "steam-turbine"
    ASSEMBLING_MACHINE_3: str = "assembling-machine-3"
    EXOSKELETON_EQUIPMENT: str = "exoskeleton-equipment"
    LONG_HANDED_INSERTER: str = "long-handed-inserter"
    ROCKET_CONTROL_UNIT: str = "rocket-control-unit"
    FAST_TRANSPORT_BELT: str = "fast-transport-belt"
    COPPER_ORE: str = "copper-ore"
    RADAR: str = "radar"
    STEEL_CHEST: str = "steel-chest"
    FILTER_INSERTER: str = "filter-inserter"
    STEEL_FURNACE: str = "steel-furnace"
    SMALL_LAMP: str = "small-lamp"
    ENERGY_SHIELD_EQUIPMENT: str = "energy-shield-equipment"
    POWER_SWITCH: str = "power-switch"
    ASSEMBLING_MACHINE_1: str = "assembling-machine-1"
    ENGINE_UNIT: str = "engine-unit"
    FUSION_REACTOR_EQUIPMENT: str = "fusion-reactor-equipment"
    RAIL_CHAIN_SIGNAL: str = "rail-chain-signal"
    PROGRAMMABLE_SPEAKER: str = "programmable-speaker"
    IRON_PLATE: str = "iron-plate"
    HEAT_EXCHANGER: str = "heat-exchanger"
    CRUDE_OIL_BARREL: str = "crude-oil-barrel"
    MEDIUM_ELECTRIC_POLE: str = "medium-electric-pole"
    URANIUM_235: str = "uranium-235"
    URANIUM_238: str = "uranium-238"
    NUCLEAR_FUEL: str = "nuclear-fuel"
    LAB: str = "lab"
    TRAIN_STOP: str = "train-stop"
    RAIL_SIGNAL: str = "rail-signal"
    COPPER_PLATE: str = "copper-plate"
    OIL_REFINERY: str = "oil-refinery"
    SPLITTER: str = "splitter"
    SOLID_FUEL: str = "solid-fuel"
    LOGISTIC_CHEST_REQUESTER: str = "logistic-chest-requester"
    LOGISTIC_CHEST_PASSIVE_PROVIDER: str = "logistic-chest-passive-provider"
    FLAMETHROWER_TURRET: str = "flamethrower-turret"
    FAST_INSERTER: str = "fast-inserter"
    ELECTRIC_FURNACE: str = "electric-furnace"
    PUMP: str = "pump"
    ELECTRIC_MINING_DRILL: str = "electric-mining-drill"
    INSERTER: str = "inserter"
    IRON_GEAR_WHEEL: str = "iron-gear-wheel"
    PIPE: str = "pipe"
    BATTERY_MK2_EQUIPMENT: str = "battery-mk2-equipment"
    EXPRESS_SPLITTER: str = "express-splitter"
    EXPRESS_TRANSPORT_BELT: str = "express-transport-belt"
    IRON_STICK: str = "iron-stick"
    LANDFILL: str = "landfill"
    LOGISTIC_CHEST_ACTIVE_PROVIDER: str = "logistic-chest-active-provider"
    OFFSHORE_PUMP: str = "offshore-pump"
    ROCKET_SILO: str = "rocket-silo"
    FAST_UNDERGROUND_BELT: str = "fast-underground-belt"
    WOOD: str = "wood"
    LAND_MINE: str = "land-mine"
    BURNER_MINING_DRILL: str = "burner-mining-drill"
    STACK_FILTER_INSERTER: str = "stack-filter-inserter"
    ASSEMBLING_MACHINE_2: str = "assembling-machine-2"
    CONCRETE: str = "concrete"
    GREEN_WIRE: str = "green-wire"
    PIPE_TO_GROUND: str = "pipe-to-ground"
    NUCLEAR_REACTOR: str = "nuclear-reactor"
    STONE_BRICK: str = "stone-brick"
    SOLAR_PANEL_EQUIPMENT: str = "solar-panel-equipment"
    LOGISTIC_ROBOT: str = "logistic-robot"
    DECIDER_COMBINATOR: str = "decider-combinator"
    STEEL_PLATE: str = "steel-plate"
    FLYING_ROBOT_FRAME: str = "flying-robot-frame"
    BEACON: str = "beacon"
    HEAVY_OIL_BARREL: str = "heavy-oil-barrel"
    LUBRICANT_BARREL: str = "lubricant-barrel"
    PROCESSING_UNIT: str = "processing-unit"
    PETROLEUM_GAS_BARREL: str = "petroleum-gas-barrel"
    TRANSPORT_BELT: str = "transport-belt"
    WATER_BARREL: str = "water-barrel"
    ARITHMETIC_COMBINATOR: str = "arithmetic-combinator"
    UNDERGROUND_BELT: str = "underground-belt"
    ELECTRONIC_CIRCUIT: str = "electronic-circuit"
    ARTILLERY_TURRET: str = "artillery-turret"
    EMPTY_BARREL: str = "empty-barrel"
    LASER_TURRET: str = "laser-turret"
    PERSONAL_ROBOPORT_EQUIPMENT: str = "personal-roboport-equipment"
    DISCHARGE_DEFENSE_EQUIPMENT: str = "discharge-defense-equipment"
    PERSONAL_LASER_DEFENSE_EQUIPMENT: str = "personal-laser-defense-equipment"
    ENERGY_SHIELD_MK2_EQUIPMENT: str = "energy-shield-mk2-equipment"
    NIGHT_VISION_EQUIPMENT: str = "night-vision-equipment"
    PERSONAL_ROBOPORT_MK2_EQUIPMENT: str = "personal-roboport-mk2-equipment"
    GUN_TURRET: str = "gun-turret"
    LOW_DENSITY_STRUCTURE: str = "low-density-structure"
    REFINED_HAZARD_CONCRETE: str = "refined-hazard-concrete"
    COAL: str = "coal"
    STORAGE_TANK: str = "storage-tank"
    BIG_ELECTRIC_POLE: str = "big-electric-pole"
    COPPER_CABLE: str = "copper-cable"
    CHEMICAL_PLANT: str = "chemical-plant"
    ELECTRIC_ENGINE_UNIT: str = "electric-engine-unit"
    STACK_INSERTER: str = "stack-inserter"
    STONE_WALL: str = "stone-wall"
    BATTERY_EQUIPMENT: str = "battery-equipment"
    STONE: str = "stone"
    BURNER_INSERTER: str = "burner-inserter"
    SOLAR_PANEL: str = "solar-panel"
    BELT_IMMUNITY_EQUIPMENT: str = "belt-immunity-equipment"
    BOILER: str = "boiler"
    STEAM_ENGINE: str = "steam-engine"
    HAZARD_CONCRETE: str = "hazard-concrete"
    HEAT_PIPE: str = "heat-pipe"
    GATE: str = "gate"
    LOGISTIC_CHEST_STORAGE: str = "logistic-chest-storage"
    ACCUMULATOR: str = "accumulator"
    STONE_FURNACE: str = "stone-furnace"
    LOGISTIC_CHEST_BUFFER: str = "logistic-chest-buffer"
    SMALL_ELECTRIC_POLE: str = "small-electric-pole"
    EXPRESS_UNDERGROUND_BELT: str = "express-underground-belt"
    ROBOPORT: str = "roboport"
    PUMPJACK: str = "pumpjack"
    SULFUR: str = "sulfur"
    PLASTIC_BAR: str = "plastic-bar"
    REFINED_CONCRETE: str = "refined-concrete"
    FAST_SPLITTER: str = "fast-splitter"
    SUBSTATION: str = "substation"
    CENTRIFUGE: str = "centrifuge"
    BATTERY: str = "battery"
    SATELLITE: str = "satellite"
    URANIUM_ORE: str = "uranium-ore"
    IRON_CHEST: str = "iron-chest"
    USED_UP_URANIUM_FUEL_CELL: str = "used-up-uranium-fuel-cell"
    TANK: str = "tank"
    SPIDERTRON: str = "spidertron"
    LOCOMOTIVE: str = "locomotive"
    CARGO_WAGON: str = "cargo-wagon"
    ARTILLERY_WAGON: str = "artillery-wagon"
    FLUID_WAGON: str = "fluid-wagon"
    CAR: str = "car"
    UTILITY_SCIENCE_PACK: str = "utility-science-pack"
    AUTOMATION_SCIENCE_PACK: str = "automation-science-pack"
    MILITARY_SCIENCE_PACK: str = "military-science-pack"
    LOGISTIC_SCIENCE_PACK: str = "logistic-science-pack"
    SPACE_SCIENCE_PACK: str = "space-science-pack"
    CHEMICAL_SCIENCE_PACK: str = "chemical-science-pack"
    PRODUCTION_SCIENCE_PACK: str = "production-science-pack"
    CANNON_SHELL: str = "cannon-shell"
    EXPLOSIVE_URANIUM_CANNON_SHELL: str = "explosive-uranium-cannon-shell"
    PIERCING_ROUNDS_MAGAZINE: str = "piercing-rounds-magazine"
    FLAMETHROWER_AMMO: str = "flamethrower-ammo"
    PIERCING_SHOTGUN_SHELL: str = "piercing-shotgun-shell"
    ROCKET: str = "rocket"
    URANIUM_CANNON_SHELL: str = "uranium-cannon-shell"
    SHOTGUN_SHELL: str = "shotgun-shell"
    EXPLOSIVE_ROCKET: str = "explosive-rocket"
    EXPLOSIVE_CANNON_SHELL: str = "explosive-cannon-shell"
    ARTILLERY_SHELL: str = "artillery-shell"
    ATOMIC_BOMB: str = "atomic-bomb"
    URANIUM_ROUNDS_MAGAZINE: str = "uranium-rounds-magazine"
    FIREARM_MAGAZINE: str = "firearm-magazine"
    EFFECTIVITY_MODULE_2: str = "effectivity-module-2"
    EFFECTIVITY_MODULE_3: str = "effectivity-module-3"
    SPEED_MODULE: str = "speed-module"
    PRODUCTIVITY_MODULE: str = "productivity-module"
    PRODUCTIVITY_MODULE_2: str = "productivity-module-2"
    EFFECTIVITY_MODULE: str = "effectivity-module"
    SPEED_MODULE_2: str = "speed-module-2"
    SPEED_MODULE_3: str = "speed-module-3"
    PRODUCTIVITY_MODULE_3: str = "productivity-module-3"
    POWER_ARMOR: str = "power-armor"
    MODULAR_ARMOR: str = "modular-armor"
    HEAVY_ARMOR: str = "heavy-armor"
    POWER_ARMOR_MK2: str = "power-armor-mk2"
    LIGHT_ARMOR: str = "light-armor"
    COMBAT_SHOTGUN: str = "combat-shotgun"
    PISTOL: str = "pistol"
    ROCKET_LAUNCHER: str = "rocket-launcher"
    SUBMACHINE_GUN: str = "submachine-gun"
    SHOTGUN: str = "shotgun"
    FLAMETHROWER: str = "flamethrower"
    CLIFF_EXPLOSIVES: str = "cliff-explosives"
    DEFENDER_CAPSULE: str = "defender-capsule"
    DISCHARGE_DEFENSE_REMOTE: str = "discharge-defense-remote"
    GRENADE: str = "grenade"
    DESTROYER_CAPSULE: str = "destroyer-capsule"
    SLOWDOWN_CAPSULE: str = "slowdown-capsule"
    ARTILLERY_TARGETING_REMOTE: str = "artillery-targeting-remote"
    DISTRACTOR_CAPSULE: str = "distractor-capsule"
    RAW_FISH: str = "raw-fish"
    POISON_CAPSULE: str = "poison-capsule"
    CLUSTER_GRENADE: str = "cluster-grenade"
    BLUEPRINT: str = "blueprint"
    BLUEPRINT_BOOK: str = "blueprint-book"
    UPGRADE_PLANNER: str = "upgrade-planner"
    DECONSTRUCTION_PLANNER: str = "deconstruction-planner"
    SPIDERTRON_REMOTE: str = "spidertron-remote"
    REPAIR_PACK: str = "repair-pack"
    RAIL: str = "rail"
    LIGHT_OIL: str = "light-oil"
    SULFURIC_ACID: str = "sulfuric-acid"
    WATER: str = "water"
    PETROLEUM_GAS: str = "petroleum-gas"
    CRUDE_OIL: str = "crude-oil"
    STEAM: str = "steam"
    LUBRICANT: str = "lubricant"
    HEAVY_OIL: str = "heavy-oil"
    SIGNAL_INFO: str = "signal-info"
    SIGNAL_DOT: str = "signal-dot"
    SIGNAL_GREY: str = "signal-grey"
    SIGNAL_5: str = "signal-5"
    SIGNAL_3: str = "signal-3"
    SIGNAL_K: str = "signal-K"
    SIGNAL_W: str = "signal-W"
    SIGNAL_YELLOW: str = "signal-yellow"
    SIGNAL_GREEN: str = "signal-green"
    SIGNAL_0: str = "signal-0"
    SIGNAL_D: str = "signal-D"
    SIGNAL_Q: str = "signal-Q"
    SIGNAL_A: str = "signal-A"
    SIGNAL_1: str = "signal-1"
    SIGNAL_4: str = "signal-4"
    SIGNAL_8: str = "signal-8"
    SIGNAL_Z: str = "signal-Z"
    SIGNAL_6: str = "signal-6"
    SIGNAL_Y: str = "signal-Y"
    SIGNAL_I: str = "signal-I"
    SIGNAL_2: str = "signal-2"
    SIGNAL_B: str = "signal-B"
    SIGNAL_X: str = "signal-X"
    SIGNAL_V: str = "signal-V"
    SIGNAL_S: str = "signal-S"
    SIGNAL_R: str = "signal-R"
    SIGNAL_G: str = "signal-G"
    SIGNAL_BLUE: str = "signal-blue"
    SIGNAL_N: str = "signal-N"
    SIGNAL_P: str = "signal-P"
    SIGNAL_T: str = "signal-T"
    SIGNAL_O: str = "signal-O"
    SIGNAL_U: str = "signal-U"
    SIGNAL_F: str = "signal-F"
    SIGNAL_M: str = "signal-M"
    SIGNAL_7: str = "signal-7"
    SIGNAL_CYAN: str = "signal-cyan"
    SIGNAL_L: str = "signal-L"
    SIGNAL_J: str = "signal-J"
    SIGNAL_H: str = "signal-H"
    SIGNAL_WHITE: str = "signal-white"
    SIGNAL_CHECK: str = "signal-check"
    SIGNAL_BLACK: str = "signal-black"
    SIGNAL_9: str = "signal-9"
    SIGNAL_PINK: str = "signal-pink"
    SIGNAL_RED: str = "signal-red"
    SIGNAL_C: str = "signal-C"
    SIGNAL_E: str = "signal-E"
    SIGNAL_EVERYTHING: str = "signal-everything"
    SIGNAL_ANYTHING: str = "signal-anything"
    SIGNAL_EACH: str = "signal-each"


WildcardSignal: List[Signal] = [
    Signal.SIGNAL_EVERYTHING,
    Signal.SIGNAL_ANYTHING,
    Signal.SIGNAL_EACH,
]

ConcreteSignal: List[Signal] = [
    signal for signal in Signal.__members__.values() if signal not in WildcardSignal
]

Signals = Dict[Signal, np.int32]



from typing import Optional

from .ASTNode import ASTNode
from .DoubleNetwork import DoubleNetwork
from .Signal import Signals, WildcardSignal


class Constant(ASTNode):
    def __init__(
        self,
        output_network: Optional[DoubleNetwork] = None,
        signals: Optional[Signals] = None,
        is_on: bool = True,
        nid: Optional[int] = None,
    ):
        super().__init__("Constant", output_network, nid)
        self.signals = signals if signals else {}
        self.is_on: bool = is_on

    @property
    def signals(self) -> Signals:
        return self._signals

    @signals.setter
    def signals(self, signal: Signals) -> None:
        assert all(
            sig not in signal.keys() for sig in WildcardSignal
        ), "can not set constant to wildcard signal"
        self._signals = signal

    def tick(self) -> None:
        self.previous_result = {}
        if self.is_on:
            self.previous_result = self.signals
            if self.output_network:
                self.output_network.update_signals(self.previous_result)

    def cli_repr(self) -> str:
        return (
            f"Constant "
            f"output={self.output_network.cli_repr() if self.output_network else '-'} "
            f"[{', '.join(f'{sig.value}: {value}' for sig, value in self.signals.items())}] "
            f"({'on' if self.is_on else 'off'})"
        )



import unittest

from factorio_circuit_ast.Operand import *


class TestSignalOperand(unittest.TestCase):
    def test_constructor(self):
        sig = Signal.SIGNAL_A
        op = SignalOperand(sig)
        self.assertEqual(op.signal, sig)

    def test_eq(self):
        op1 = SignalOperand(Signal.SIGNAL_A)
        op2 = SignalOperand(Signal.SIGNAL_B)
        self.assertNotEqual(op1, op2)
        op2.signal = Signal.SIGNAL_A
        self.assertNotEqual(op1, None)
        self.assertEqual(op1, op2)

    def test_abstract(self):
        op = SignalOperand(Signal.SIGNAL_A)
        self.assertEqual(op.wildcard(), False)
        op.signal = Signal.SIGNAL_EACH
        self.assertEqual(op.wildcard(), True)
        op.signal = Signal.SIGNAL_EVERYTHING
        self.assertEqual(op.wildcard(), True)
        op.signal = Signal.SIGNAL_ANYTHING
        self.assertEqual(op.wildcard(), True)
        op.signal = Signal.WATER
        self.assertEqual(op.wildcard(), False)

    def test_each(self):
        op = SignalOperand(Signal.SIGNAL_A)
        self.assertEqual(op.is_each(), False)
        op.signal = Signal.SIGNAL_EACH
        self.assertEqual(op.is_each(), True)
        op.signal = Signal.SIGNAL_EVERYTHING
        self.assertEqual(op.is_each(), False)
        op.signal = Signal.SIGNAL_ANYTHING
        self.assertEqual(op.is_each(), False)
        op.signal = Signal.WATER
        self.assertEqual(op.is_each(), False)

    def test_anything(self):
        op = SignalOperand(Signal.SIGNAL_A)
        self.assertEqual(op.is_anything(), False)
        op.signal = Signal.SIGNAL_EACH
        self.assertEqual(op.is_anything(), False)
        op.signal = Signal.SIGNAL_EVERYTHING
        self.assertEqual(op.is_anything(), False)
        op.signal = Signal.SIGNAL_ANYTHING
        self.assertEqual(op.is_anything(), True)
        op.signal = Signal.WATER
        self.assertEqual(op.is_anything(), False)

    def test_everything(self):
        op = SignalOperand(Signal.SIGNAL_A)
        self.assertEqual(op.is_everything(), False)
        op.signal = Signal.SIGNAL_EACH
        self.assertEqual(op.is_everything(), False)
        op.signal = Signal.SIGNAL_EVERYTHING
        self.assertEqual(op.is_everything(), True)
        op.signal = Signal.SIGNAL_ANYTHING
        self.assertEqual(op.is_everything(), False)
        op.signal = Signal.WATER
        self.assertEqual(op.is_everything(), False)


class TestConstantOperand(unittest.TestCase):
    def test_constructo(self):
        op = ConstantOperand(np.int32(10))
        self.assertEqual(op.constant, 10)

    def test_eq(self):
        op1 = ConstantOperand(np.int32(10))
        op2 = ConstantOperand(np.int32(20))
        self.assertNotEqual(op1, op2)
        op3 = SignalOperand(Signal.SIGNAL_A)
        self.assertNotEqual(op1, op3)
        op2.constant = 10
        self.assertEqual(op1, op2)

    def test_abstract(self):
        op = ConstantOperand(np.int32(12))
        self.assertEqual(op.wildcard(), False)

    def test_each(self):
        op = ConstantOperand(np.int32(12))
        self.assertEqual(op.is_each(), False)

    def test_anything(self):
        op = ConstantOperand(np.int32(12))
        self.assertEqual(op.is_anything(), False)

    def test_everything(self):
        op = ConstantOperand(np.int32(12))
        self.assertEqual(op.is_everything(), False)



from typing import Optional, Any

import numpy as np

from .Network import Network, NetworkType
from .Signal import Signal, WildcardSignal, Signals


class DoubleNetwork:
    def __init__(
        self, red: Optional[Network] = None, green: Optional[Network] = None
    ) -> None:
        self.red = red
        self.green = green

    def get_signal(self, signal: Signal) -> np.int32:
        assert signal not in WildcardSignal
        return (self.red.get_signal(signal) if self.red else np.int32(0)) + (
            self.green.get_signal(signal) if self.green else np.int32(0)
        )

    def get_signals(self) -> Signals:
        result: Signals = self.red.get_signals() if self.red else {}
        if self.green:
            for signal, value in self.green.get_signals().items():
                result[signal] = result.setdefault(signal, np.int32(0)) + value
        return result

    def update_signal(self, signal: Signal, value: np.int32) -> None:
        assert signal not in WildcardSignal
        if self.red:
            self.red.update_signal(signal, value)
        if self.green:
            self.green.update_signal(signal, value)

    def update_signals(self, signals: Signals) -> None:
        assert (
            Signal.SIGNAL_EACH not in signals
            and Signal.SIGNAL_EVERYTHING not in signals
            and Signal.SIGNAL_ANYTHING not in signals
        ), "only concrete signals are allowed in update_signals"
        if self.red:
            self.red.update_signals(signals)
        if self.green:
            self.green.update_signals(signals)

    @property
    def red(self) -> Optional[Network]:
        return self._red

    @red.setter
    def red(self, net: Optional[Network]) -> None:
        if net and net.type != NetworkType.RED:
            raise TypeError("Can not assign green network to red DoubleNetwork part")
        self._red = net

    @property
    def green(self) -> Optional[Network]:
        return self._green

    @green.setter
    def green(self, net: Optional[Network]) -> None:
        if net and net.type != NetworkType.GREEN:
            raise TypeError("Can not assign green network to red DoubleNetwork part")
        self._green = net

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DoubleNetwork):
            return self.red == other.red and self.green == other.green
        return False

    def __repr__(self) -> str:
        return f"DoubleNetwork(red={self.red!r}, green={self.green!r})"

    def cli_repr(self) -> str:
        return f"[r={self.red.cli_repr() if self.red else '-'},g={self.green.cli_repr() if self.green else '-'}]"



from __future__ import annotations

from enum import Enum
from typing import List, Dict, Optional
from weakref import ref, ReferenceType

import numpy as np

from .ASTNode import ASTNode
from .Signal import Signal, WildcardSignal, Signals

networks: Dict[int, Network] = {}


class NetworkType(Enum):
    RED = "red"
    GREEN = "green"


class Network:
    def __init__(
        self, nid: Optional[int] = None, network: Optional[NetworkType] = None
    ) -> None:
        self.type: NetworkType = network if network else NetworkType.RED
        self._depends: List[ReferenceType[ASTNode]] = []
        self._dependants: List[ReferenceType[ASTNode]] = []
        self._previous_state: Signals = {}
        self._current_state: Signals = {}
        self.id = nid if nid is not None else max([*networks.keys(), 0]) + 1
        networks[self.id] = self

    def depends_on(self, node: ASTNode) -> None:
        self._depends.append(ref(node))

    def dependant_from(self, node: ASTNode) -> None:
        self._dependants.append(ref(node))

    def get_signal(self, signal: Signal) -> np.int32:
        assert (
            signal not in WildcardSignal
        ), f"only concrete signals are allowed in get_signal, got {signal}"
        return self._previous_state.setdefault(signal, np.int32(0))

    def get_signals(self) -> Signals:
        return self._previous_state

    def tick(self) -> None:
        self._previous_state = self._current_state
        self._current_state = {}

    def update_signal(self, signal: Signal, value: np.int32) -> None:
        assert (
            signal not in WildcardSignal
        ), f"only concrete signals are allowed in update_signal, got {signal}"
        self._current_state[signal] = (
            self._current_state.setdefault(signal, np.int32(0)) + value
        )

    def update_signals(self, signals: Signals) -> None:
        assert (
            Signal.SIGNAL_EACH not in signals
            and Signal.SIGNAL_EVERYTHING not in signals
            and Signal.SIGNAL_ANYTHING not in signals
        ), "only concrete signals are allowed in update_signals"
        self._current_state.update(signals)

    @property
    def depends(self) -> List[ReferenceType[ASTNode]]:
        return self._depends

    @property
    def dependants(self) -> List[ReferenceType[ASTNode]]:
        return self._dependants

    def __repr__(self) -> str:
        return f"Network(nid={self.id!r}, network={self.type!r})"

    def cli_repr(self) -> str:
        return f"n{self.id}"


def get_or_create_network(id: int, type: NetworkType) -> Network:
    if id in networks:
        assert (
            net := networks[id]
        ).type == type, "The stored network has the wrong type"
        return net
    return Network(id, type)



from enum import Enum
from typing import Optional, Any

import numpy as np

from .DoubleNetwork import DoubleNetwork
from .Operand import Operand, ConstantOperand, SignalOperand
from .Signal import Signal, Signals
from .TwoSidedASTNode import TwoSidedASTNode


class NumericOperator(Enum):
    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    MODULO = "%"
    EXPONENT = "^"
    LEFT_BIT_SHIFT = "<<"
    RIGHT_BIT_SHIFT = ">>"
    AND = "AND"
    OR = "OR"
    XOR = "XOR"

    def calculate(self, left: np.int32, right: np.int32) -> np.int32:
        if self == NumericOperator.ADD:
            return left + right
        elif self == NumericOperator.SUBTRACT:
            return left - right
        elif self == NumericOperator.MULTIPLY:
            return left * right
        elif self == NumericOperator.DIVIDE:
            return left // right
        elif self == NumericOperator.MODULO:
            return left % right
        elif self == NumericOperator.EXPONENT:
            return left**right
        elif self == NumericOperator.LEFT_BIT_SHIFT:
            return left << right
        elif self == NumericOperator.RIGHT_BIT_SHIFT:
            return left >> right
        elif self == NumericOperator.AND:
            return left & right
        elif self == NumericOperator.OR:
            return left | right
        elif self == NumericOperator.XOR:
            return left ^ right
        raise NotImplementedError(f"Unknown operator {self}")


class Operation(TwoSidedASTNode):
    def __init__(
        self,
        operation: NumericOperator,
        result: SignalOperand,
        output_network: Optional[DoubleNetwork] = None,
        input_network: Optional[DoubleNetwork] = None,
        left: Optional[Operand] = None,
        right: Optional[Operand] = None,
        nid: Optional[int] = None,
    ):
        super().__init__("Operation", output_network, input_network, nid)
        self.operation: NumericOperator = operation
        # set these for comparison
        self._left: Operand = left if left else ConstantOperand(np.int32(0))
        self._right: Operand = right if right else ConstantOperand(np.int32(0))
        self._result: SignalOperand = result
        # and set them with checks
        self.result = result
        self.left = self._left
        self.right = self._right

    @property
    def left(self) -> Operand:
        return self._left

    @left.setter
    def left(self, left: Operand) -> None:
        if left and isinstance(left, SignalOperand):
            assert self.input_network, "can not set signal without network"
        if left.is_each():
            assert (
                not self.right.is_each()
            ), "Only one operand for Operation may be Each"
        elif left.wildcard():
            raise AssertionError(
                "Only concrete signals or Each are allowed in Operation"
            )
        else:
            assert (
                not self.result.is_each() or self.right.is_each()
            ), "Can not assign signal other than each to Operation while having result each"
        self._left = left

    @property
    def right(self) -> Operand:
        return self._right

    @right.setter
    def right(self, right: Operand) -> None:
        if right and isinstance(right, SignalOperand):
            assert self.input_network, "can not set signal without network"
        if right.is_each():
            assert not self.left.is_each(), "Only one operand for Operation may be Each"
        elif right.wildcard():
            raise AssertionError(
                "Only concrete signals or Each are allowed in Operation"
            )
        else:
            assert (
                not self.result.is_each() or self.left.is_each()
            ), "Can not assign signal other than each to Operation while having result each"
        self._right = right

    @property
    def result(self) -> SignalOperand:
        return self._result

    @result.setter
    def result(self, op: SignalOperand) -> None:
        assert op.signal not in (
            Signal.SIGNAL_EVERYTHING,
            Signal.SIGNAL_ANYTHING,
        ), "Anything or Everything are not allowed as Operation results"
        if op.is_each():
            assert (
                self.left.is_each() or self.right.is_each()
            ), "Need to have one Operand as Each in order to set result to each"
        self._result = op

    def tick(self) -> None:
        results: Signals = {}
        if not (self.left.is_each() or self.right.is_each()):
            assert (
                not self.result.is_each()
            ), "result is each, but neither left nor right are each"
            results[self.result.signal] = self.operation.calculate(
                self._get_input_value(self.left),
                self._get_input_value(self.right),
            )
        else:
            assert self.input_network
            if self.left.is_each():
                assert self.input_network
                for signal, value in self.input_network.get_signals().items():
                    results[signal] = self.operation.calculate(
                        value, self._get_input_value(self.right)
                    )
            elif self.right.is_each():
                for signal, value in self.input_network.get_signals().items():
                    results[signal] = self.operation.calculate(
                        self._get_input_value(self.left), value
                    )
            else:
                raise NotImplementedError("unreachable")
        if self.output_network:
            if self.result.is_each():
                for signal, value in results.items():
                    self.output_network.update_signal(signal, value)
            else:
                self.output_network.update_signal(
                    self.result.signal, np.sum([*results.values()], dtype=np.int32)
                )
        self.previous_result = results

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Operation):
            return (
                super().__eq__(other)
                and self.operation == other.operation
                and self.right == other.right
                and self.left == other.left
                and self.result == other.result
            )
        return False

    def __repr__(self) -> str:
        return (
            f"Operation("
            f"operation={self.operation}, "
            f"result={self.result}, "
            f"output_network={self.output_network}, "
            f"input_network={self.input_network}, "
            f"left={self.left}, "
            f"right={self.right})"
        )

    def cli_repr(self) -> str:
        return (
            f"Operation "
            f"input={self.input_network.cli_repr() if self.input_network else '-'} "
            f"output={self.output_network.cli_repr() if self.output_network else '-'} "
            f"({self.left.cli_repr()} {self.operation.value} {self.right.cli_repr()}) "
            f"-> {self.result.cli_repr()}"
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, TYPE_CHECKING, Any, Dict

from .Signal import Signals

if TYPE_CHECKING:
    from .DoubleNetwork import DoubleNetwork


nodes: Dict[int, ASTNode] = {}


class ASTNode(ABC):
    def __init__(
        self,
        name: str,
        output_network: Optional[DoubleNetwork] = None,
        nid: Optional[int] = None,
    ):
        self.name: str = name
        self.output_network: Optional[DoubleNetwork] = output_network
        self.previous_result: Optional[Signals] = None
        self.id = nid if nid is not None else max([*nodes.keys(), 0]) + 1
        nodes[self.id] = self

    @abstractmethod
    def tick(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def cli_repr(self) -> str:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ASTNode):
            return (
                self.name == other.name and self.output_network == other.output_network
            )
        return False  # pragma: no cover



from enum import Enum
from typing import Optional, Any

import numpy as np

from .DoubleNetwork import DoubleNetwork
from .Operand import Operand, SignalOperand, ConstantOperand
from .Signal import Signal
from .TwoSidedASTNode import TwoSidedASTNode


class DeciderOperator(Enum):
    GREATER_THAN: str = ">"
    LESS_THAN: str = "<"
    GREATER_OR_EQUAL: str = ">="
    LESS_OR_EQUAL: str = "<="
    EQUAL: str = "="
    NOT_EQUAL: str = "!="

    def calculate(self, left: np.int32, right: np.int32) -> np.int32:
        result: np.bool_
        if self == DeciderOperator.GREATER_THAN:
            result = left > right
        elif self == DeciderOperator.LESS_THAN:
            result = left < right
        elif self == DeciderOperator.GREATER_OR_EQUAL:
            result = left >= right
        elif self == DeciderOperator.LESS_OR_EQUAL:
            result = left <= right
        elif self == DeciderOperator.EQUAL:
            result = left == right
        elif self == DeciderOperator.NOT_EQUAL:
            result = left != right
        else:
            raise NotImplementedError(f"Unknown operator {self}")
        return np.int32(result)


class Comparison(TwoSidedASTNode):
    def __init__(
        self,
        operation: DeciderOperator,
        result: SignalOperand,
        output_network: Optional[DoubleNetwork] = None,
        input_network: Optional[DoubleNetwork] = None,
        left: Optional[Operand] = None,
        right: Optional[Operand] = None,
        copy_count_from_input: bool = True,
        nid: Optional[int] = None,
    ):
        super().__init__("Comparison", output_network, input_network, nid)
        self.operation: DeciderOperator = operation
        self._result: SignalOperand = result
        self._left: Operand = left if left else ConstantOperand(np.int32(0))
        self._right: Operand = right if right else ConstantOperand(np.int32(0))
        self.result = result
        self.left = self._left
        self.right = self._right
        self.copy_count_from_input: bool = copy_count_from_input

    @property
    def left(self) -> Operand:
        return self._left

    @left.setter
    def left(self, left: Operand) -> None:
        if left and isinstance(left, SignalOperand):
            assert self.input_network, "can not set signal without network"
        if left.wildcard():
            if left.is_everything():
                assert not (
                    self.result.is_each() or self.result.is_anything()
                ), "everything wildcard needs everything concrete value as result"
            elif left.is_anything():
                assert (
                    not self.result.is_each()
                ), "anything wildcard needs anything, everything or concrete signal as result"
            elif left.is_each():
                assert not (
                    self.result.is_anything() or self.result.is_everything()
                ), "each wildcard needs each or concrete signal as result"
            else:
                raise NotImplementedError(f"Unknown wildcard signal {left}")
        else:
            assert not (
                self.result.is_each() or self.result.is_anything()
            ), "concrete signal needs everything or concrete signal as result"
        self._left = left

    @property
    def right(self) -> Operand:
        return self._right

    @right.setter
    def right(self, right: Operand) -> None:
        assert (
            not right.wildcard()
        ), "right side of comparison is not allowed to be a wildcard"
        if right and isinstance(right, SignalOperand):
            assert self.input_network, "can not set signal without network"
        self._right = right

    @property
    def result(self) -> SignalOperand:
        return self._result

    @result.setter
    def result(self, op: SignalOperand) -> None:
        if op.wildcard():
            assert (
                self.input_network
            ), "can not use result wildcards without input network"
            if op.is_everything():
                assert (
                    not self.left.is_each()
                ), "everything wildcard may not be the result of each"
            elif op.is_anything():
                assert (
                    self.left.is_anything()
                ), "anything wildcard needs anything signal on the left side"
            elif op.is_each():
                assert (
                    self.left.is_each()
                ), "each wildcard needs each signal on the left side"
            else:
                raise NotImplementedError(f"Unknown wildcard signal {op}")
        self._result = op

    def tick(self) -> None:
        self.previous_result = {}
        right_value: np.int32 = self._get_input_value(self.right)
        if self.left.is_each():
            assert self.input_network
            for key, value in self.input_network.get_signals().items():
                if self.operation.calculate(value, right_value):
                    self.previous_result[key] = (
                        value if self.copy_count_from_input else np.int32(1)
                    )
            if not self.result.is_each() and self.previous_result:
                self.previous_result = {
                    self.result.signal: np.sum(
                        [*self.previous_result.values()], dtype=np.int32
                    )
                }
        else:
            passed: bool = False
            right_signal: Optional[Signal] = (
                right.signal if isinstance(right := self.right, SignalOperand) else None
            )
            if self.left.is_anything():
                assert self.input_network
                for key, value in self.input_network.get_signals().items():
                    if right_signal and key == right_signal:
                        continue
                    if self.operation.calculate(value, right_value):
                        self.previous_result = {
                            key: value if self.copy_count_from_input else np.int32(1)
                        }
                        passed = True
                        break
            elif self.left.is_everything():
                assert self.input_network
                passed = True
                for key, value in self.input_network.get_signals().items():
                    if right_signal and key == right_signal:
                        continue
                    if not self.operation.calculate(value, right_value):
                        passed = False
                        break
            else:
                passed = bool(
                    self.operation.calculate(
                        self._get_input_value(self.left),
                        right_value,
                    )
                )
            if passed and self.input_network:
                if self.result.is_everything():
                    self.previous_result = {
                        key: value if self.copy_count_from_input else np.int32(1)
                        for key, value in self.input_network.get_signals().items()
                        if value != 0
                    }
                elif not self.result.is_anything():
                    self.previous_result = {
                        self.result.signal: self._get_input_value(self.result)
                        if self.copy_count_from_input
                        else np.int32(1)
                    }
            elif passed:
                if not self.result.is_everything() and not self.copy_count_from_input:
                    self.previous_result = {self.result.signal: np.int32(1)}
        if self.output_network:
            self.output_network.update_signals(self.previous_result)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Comparison):
            return (
                super().__eq__(other)
                and self.operation == other.operation
                and self.right == other.right
                and self.left == other.left
                and self.result == other.result
                and self.copy_count_from_input == other.copy_count_from_input
            )
        return False

    def __repr__(self) -> str:
        return (
            f"Comparison("
            f"operation={self.operation}, "
            f"result={self.result}, "
            f"output_network={self.output_network}, "
            f"input_network={self.input_network}, "
            f"left={self.left}, "
            f"right={self.right}, "
            f"copy_count_from_input={self.copy_count_from_input})"
        )

    def cli_repr(self) -> str:
        return (
            f"Comparison "
            f"input={self.input_network.cli_repr() if self.input_network else '-'} "
            f"output={self.output_network.cli_repr() if self.output_network else '-'} "
            f"({self.left.cli_repr()} {self.operation.value} {self.right.cli_repr()}) "
            f"-> {self.result.cli_repr()} "
            f"({'copy' if self.copy_count_from_input else 'no_copy'})"
        )



import unittest

from factorio_circuit_ast.Comparison import *
from factorio_circuit_ast.Network import Network, NetworkType
from factorio_circuit_ast.Signal import Signal, Signals


class TestDeciderOperator(unittest.TestCase):
    def test_calculate_greater_than(self):
        res = DeciderOperator.GREATER_THAN.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, False)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.GREATER_THAN.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, False)

    def test_calculate_less_than(self):
        res = DeciderOperator.LESS_THAN.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, True)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.LESS_THAN.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, False)

    def test_calculate_greater_or_equal(self):
        res = DeciderOperator.GREATER_OR_EQUAL.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, False)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.GREATER_OR_EQUAL.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, True)

    def test_calculate_less_or_equal(self):
        res = DeciderOperator.LESS_OR_EQUAL.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, True)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.LESS_OR_EQUAL.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, True)

    def test_calculate_equal(self):
        res = DeciderOperator.EQUAL.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, False)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.EQUAL.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, True)

    def test_calculate_not_equal(self):
        res = DeciderOperator.NOT_EQUAL.calculate(np.int32(10), np.int32(20))
        self.assertEqual(res, True)
        self.assertIsInstance(res, np.int32)
        res = DeciderOperator.NOT_EQUAL.calculate(np.int32(10), np.int32(10))
        self.assertEqual(res, False)


def get_nets() -> DoubleNetwork:
    r_net = Network(network=NetworkType.RED)
    g_net = Network(network=NetworkType.GREEN)
    d_net = DoubleNetwork(r_net, g_net)
    return d_net


class TestComparison(unittest.TestCase):
    def test_constructor(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
        )
        self.assertEqual(cmp.result, SignalOperand(Signal.SIGNAL_A))
        self.assertEqual(cmp.operation, DeciderOperator.EQUAL)
        self.assertEqual(cmp.left, ConstantOperand(np.int32(0)))
        self.assertEqual(cmp.right, ConstantOperand(np.int32(0)))
        self.assertEqual(cmp.copy_count_from_input, True)
        self.assertEqual(cmp.input_network, None)
        self.assertEqual(cmp.output_network, None)
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            i_net,
            SignalOperand(Signal.SIGNAL_RED),
            SignalOperand(Signal.SIGNAL_R),
            False,
        )
        self.assertEqual(cmp.result, SignalOperand(Signal.SIGNAL_A))
        self.assertEqual(cmp.operation, DeciderOperator.EQUAL)
        self.assertEqual(cmp.left, SignalOperand(Signal.SIGNAL_RED))
        self.assertEqual(cmp.right, SignalOperand(Signal.SIGNAL_R))
        self.assertEqual(cmp.copy_count_from_input, False)
        self.assertIs(cmp.input_network, i_net)
        self.assertIs(cmp.output_network, o_net)

    @staticmethod
    def setter_left_auto(input_o: Operand, output_o: SignalOperand):
        i_net = get_nets()
        o_net = get_nets()
        Comparison(
            DeciderOperator.EQUAL,
            output_o,
            o_net,
            i_net,
            left=input_o,
        )

    @staticmethod
    def setter_right_auto(input_o: Operand, output_o: SignalOperand):
        i_net = get_nets()
        o_net = get_nets()
        Comparison(
            DeciderOperator.EQUAL,
            output_o,
            o_net,
            i_net,
            right=input_o,
        )

    @staticmethod
    def setter_result_auto(input_o: Operand, output_o: SignalOperand):
        i_net = get_nets()
        o_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            i_net,
            input_o,
        )
        cmp.result = output_o

    def setter_test_auto(self, input_o: Operand, output: Signal, is_ok: bool):
        output_o = SignalOperand(output)
        if not is_ok:
            with self.assertRaises(AssertionError):
                self.setter_left_auto(input_o, output_o)
            with self.assertRaises(AssertionError):
                self.setter_right_auto(input_o, output_o)
            with self.assertRaises(AssertionError):
                self.setter_result_auto(input_o, output_o)
        else:
            self.setter_left_auto(input_o, output_o)
            if input_o.wildcard():
                with self.assertRaises(AssertionError):
                    self.setter_right_auto(input_o, output_o)
            else:
                self.setter_right_auto(input_o, output_o)
            self.setter_result_auto(input_o, output_o)

    def test_setter_output_everything(self):
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EACH), Signal.SIGNAL_EVERYTHING, False
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_ANYTHING), Signal.SIGNAL_EVERYTHING, True
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EVERYTHING), Signal.SIGNAL_EVERYTHING, True
        )
        self.setter_test_auto(
            ConstantOperand(np.int32(10)), Signal.SIGNAL_EVERYTHING, True
        )

    def test_setter_output_anything(self):
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EACH), Signal.SIGNAL_ANYTHING, False
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EVERYTHING), Signal.SIGNAL_ANYTHING, False
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_ANYTHING), Signal.SIGNAL_ANYTHING, True
        )
        self.setter_test_auto(
            ConstantOperand(np.int32(10)), Signal.SIGNAL_ANYTHING, False
        )

    def test_setter_output_each(self):
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EACH), Signal.SIGNAL_EACH, True
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EVERYTHING), Signal.SIGNAL_EACH, False
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_ANYTHING), Signal.SIGNAL_EACH, False
        )
        self.setter_test_auto(ConstantOperand(np.int32(10)), Signal.SIGNAL_EACH, False)

    def test_setter_output_normal(self):
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EACH), Signal.ACCUMULATOR, True
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_EVERYTHING), Signal.ACCUMULATOR, True
        )
        self.setter_test_auto(
            SignalOperand(Signal.SIGNAL_ANYTHING), Signal.ACCUMULATOR, True
        )
        self.setter_test_auto(ConstantOperand(np.int32(10)), Signal.ACCUMULATOR, True)

    def test_left_setter_no_network(self):
        with self.assertRaises(AssertionError):
            Comparison(
                DeciderOperator.EQUAL,
                SignalOperand(Signal.SIGNAL_A),
                left=SignalOperand(Signal.SIGNAL_RED),
            )

    def test_right_setter_no_network(self):
        with self.assertRaises(AssertionError):
            Comparison(
                DeciderOperator.EQUAL,
                SignalOperand(Signal.SIGNAL_A),
                right=SignalOperand(Signal.SIGNAL_A),
            )

    def test_result_setter_no_network(self):
        with self.assertRaises(AssertionError):
            Comparison(DeciderOperator.EQUAL, SignalOperand(Signal.SIGNAL_EVERYTHING))

    def test_eq(self):
        i_net = get_nets()
        o_net = get_nets()
        op1 = Comparison(DeciderOperator.EQUAL, SignalOperand(Signal.SIGNAL_A))
        op2 = Comparison(DeciderOperator.NOT_EQUAL, SignalOperand(Signal.SIGNAL_A))
        self.assertNotEqual(op1, ConstantOperand(np.int32(2)))
        self.assertNotEqual(op1, op2)
        op2.operation = DeciderOperator.EQUAL
        op1.left = ConstantOperand(np.int32(10))
        self.assertNotEqual(op1, op2)
        op1.left = ConstantOperand(np.int32(0))
        op2.right = ConstantOperand(np.int32(50))
        self.assertNotEqual(op1, op2)
        op2.right = ConstantOperand(np.int32(0))
        op1.result = SignalOperand(Signal.SIGNAL_B)
        self.assertNotEqual(op1, op2)
        op1.result = SignalOperand(Signal.SIGNAL_A)
        op2.output_network = o_net
        self.assertNotEqual(op1, op2)
        op1.output_network = o_net
        op1.input_network = i_net
        self.assertNotEqual(op1, op2)
        op2.input_network = i_net
        op2.copy_count_from_input = False
        self.assertNotEqual(op1, op2)
        op1.copy_count_from_input = False
        self.assertEqual(op1, op2)

    def test_previous_result_after_tick(self):
        o_net = get_nets()
        cmp = Comparison(DeciderOperator.EQUAL, SignalOperand(Signal.SIGNAL_A), o_net)
        cmp.tick()
        o_net.red.tick()
        self.assertEqual(cmp.previous_result, o_net.get_signals())

    def test_compare_constant(self):
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
            left=ConstantOperand(np.int32(10)),
            right=ConstantOperand(np.int32(20)),
            copy_count_from_input=False,
        )
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})
        cmp.right = ConstantOperand(np.int32(10))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_compare_constant_result_everything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_EVERYTHING),
            o_net,
            i_net,
            left=ConstantOperand(np.int32(10)),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 20,
            Signal.SIGNAL_E: 30,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {s: 1 for s in sig.keys()})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, sig)

    def test_compare_everything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.GREATER_THAN,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_EVERYTHING),
            right=ConstantOperand(np.int32(5)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_R: 20,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        cmp.result.signal = Signal.SIGNAL_R
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 1})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 20})
        cmp.result.signal = Signal.SIGNAL_A
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 10})
        cmp.operation = DeciderOperator.EQUAL
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_compare_everything_result_everything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.GREATER_THAN,
            SignalOperand(Signal.SIGNAL_EVERYTHING),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_EVERYTHING),
            right=ConstantOperand(np.int32(5)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_R: 20,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {s: 1 for s in sig.keys()})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, sig)
        cmp.operation = DeciderOperator.EQUAL
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})
        cmp.operation = DeciderOperator.GREATER_THAN
        cmp.right = SignalOperand(Signal.SIGNAL_A)
        cmp.tick()
        self.assertEqual(cmp.previous_result, sig)

    def test_compare_anything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_ANYTHING),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_R: 20,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        cmp.result.signal = Signal.SIGNAL_R
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 1})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 20})
        cmp.right = ConstantOperand(np.int32(5))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})
        i_net.update_signal(Signal.SIGNAL_A, np.int32(10))
        i_net.red.tick()
        cmp.right = SignalOperand(Signal.SIGNAL_A)
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_compare_anything_result_anything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_ANYTHING),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_ANYTHING),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_R: 20,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        # TODO: Implement deterministic order and test order
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 10})
        cmp.right = ConstantOperand(np.int32(5))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_compare_anything_result_everything(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_EVERYTHING),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_ANYTHING),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: 10,
            Signal.SIGNAL_R: 20,
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1, Signal.SIGNAL_R: 1})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(
            cmp.previous_result, {Signal.SIGNAL_A: 10, Signal.SIGNAL_R: 20}
        )
        cmp.right = ConstantOperand(np.int32(5))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_compare_each(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_EACH),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: np.int32(10),
            Signal.SIGNAL_R: np.int32(20),
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        cmp.result.signal = Signal.SIGNAL_R
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 1})
        sig[Signal.SIGNAL_E] = np.int32(10)
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 2})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 20})
        cmp.right = ConstantOperand(np.int32(5))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})
        cmp.right = SignalOperand(Signal.SIGNAL_A)
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_R: 20})

    def test_compare_each_result_each(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_EACH),
            o_net,
            i_net,
            left=SignalOperand(Signal.SIGNAL_EACH),
            right=ConstantOperand(np.int32(10)),
            copy_count_from_input=False,
        )
        sig: Signals = {
            Signal.SIGNAL_A: np.int32(10),
            Signal.SIGNAL_R: np.int32(20),
        }
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1})
        sig[Signal.SIGNAL_E] = np.int32(10)
        i_net.update_signals(sig)
        i_net.red.tick()
        cmp.tick()
        self.assertEqual(cmp.previous_result, {Signal.SIGNAL_A: 1, Signal.SIGNAL_E: 1})
        cmp.copy_count_from_input = True
        cmp.tick()
        self.assertEqual(
            cmp.previous_result, {Signal.SIGNAL_A: 10, Signal.SIGNAL_E: 10}
        )
        cmp.right = ConstantOperand(np.int32(5))
        cmp.tick()
        self.assertEqual(cmp.previous_result, {})

    def test_cli_repr(self):
        o_net = get_nets()
        i_net = get_nets()
        cmp = Comparison(
            DeciderOperator.EQUAL,
            SignalOperand(Signal.SIGNAL_A),
        )
        self.assertEqual(
            cmp.cli_repr(), "Comparison input=- output=- (0 = 0) -> signal-A (copy)"
        )
        cmp = Comparison(
            DeciderOperator.GREATER_THAN,
            SignalOperand(Signal.SIGNAL_R),
            o_net,
            i_net,
            SignalOperand(Signal.SIGNAL_B),
            ConstantOperand(np.int32(5)),
            False,
        )
        self.assertEqual(
            cmp.cli_repr(),
            f"Comparison input={i_net.cli_repr()} output={o_net.cli_repr()} (signal-B > 5) -> signal-R (no_copy)",
        )



from abc import ABC
from typing import Optional, Any

import numpy as np

from .ASTNode import ASTNode
from .DoubleNetwork import DoubleNetwork
from .Operand import Operand, SignalOperand, ConstantOperand


class TwoSidedASTNode(ASTNode, ABC):
    def __init__(
        self,
        name: str,
        output_network: Optional[DoubleNetwork] = None,
        input_network: Optional[DoubleNetwork] = None,
        nid: Optional[int] = None,
    ):
        super().__init__(name, output_network, nid)
        self.input_network: Optional[DoubleNetwork] = input_network

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, TwoSidedASTNode):
            return super().__eq__(other) and self.input_network == other.input_network
        return False  # pragma: no cover

    def _get_input_value(self, op: Operand) -> np.int32:
        if isinstance(op, SignalOperand):
            assert self.input_network
            return self.input_network.get_signal(op.signal)
        elif isinstance(op, ConstantOperand):
            return op.constant
        raise NotImplementedError(f"Invalid operand {op}")
